/*
 * This program and all associated contents are
 * Copyright (C) 1996,1997,1998,1999,200,2001 David G. Andersen
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * */

#include <stdio.h>
#include <sys/types.h>
#include <string.h>

/*#define LOOPCOUNT*/

long int sm_search(unsigned char *buf, char *searchstring, size_t size, size_t start, int m)
{
    int i, j;
    int dig;

    for (i = start; i < size*2; i++) {
	for (j = 0; j < m; j++) {
	    dig = (unsigned char)buf[(i+j)/2];
	    if (!((i+j)%2)) { /* First digit in the byte */
		dig = (dig & 0xf0) >> 4;
	    } else {
		dig = (dig & 0x0f);
	    }
	    if (dig != searchstring[j]) { break; }
	}
	if (j == m) {
	    return i;
	}
    }

    return i;
}


size_t big_search(unsigned char *buf, unsigned char *searchstring, size_t size, size_t start, int m)
{
    int i, j;
    int dig;
    unsigned char j1, offset_j2, j2, offset_j3;
#ifdef LOOPCOUNT
    int eloop = 0, oloop = 0;
#endif

    j1 =        searchstring[0] << 4 | searchstring[1];
    offset_j2 = searchstring[1] << 4 | searchstring[2];
    j2 =        searchstring[2] << 4 | searchstring[3];
    offset_j3 = searchstring[3] << 4 | searchstring[4];

    size*=2;

    i = start;
    if (i & 1) {
	for (j = 0; j < m; j++) {
	    dig = (unsigned char)buf[(i+j)/2];
	    if ((i+j)&1) {
		dig = (dig & 0x0f);
	    } else {
		dig = (dig&0xf0) >> 4;
	    }
	    if (dig != searchstring[j]) { break; }
	}
	if (j == m) {
	    return start;
	}
	start++;
    }
    
    /* Assume m >= 4 */
    for (i = start; i < size; i++) {

	/* i = 0, 2, 4, 6 ... */
	if (buf[i/2] == j1 &&
	    buf[(i/2) + 1] == j2) {
#ifdef LOOPCOUNT
	    eloop++;
#endif
	    for (j = 4; j < m; j++) {
		dig = (unsigned char)buf[(i+j)/2];
		if ((i+j)&1) {
		    dig = (dig & 0x0f);
		} else {
		    dig = (dig&0xf0) >> 4;
		}
		if (dig != searchstring[j]) { break; }
	    }
	    if (j == m) {
#ifdef LOOPCOUNT
		printf("eloop: %d  oloop: %d\n", eloop, oloop);
#endif
		return i;
	    }
	}
	i++;

	/* i = 1, 3, 5, 7 ... */

	if (((buf[i/2] & 0x0f) == ((j1 & 0xf0)>>4))
	    && (buf[((i+1)/2)] == offset_j2)
	    && (buf[((i+3)/2)] == offset_j3)
	    ) {
	    for (j = 5; j < m; j++) {
#ifdef LOOPCOUNT
		oloop++;
#endif
		dig = (unsigned char)buf[(i+j)/2];
		if ((i+j)&1) {
		    dig = (dig & 0x0f);
		} else {
		    dig = (dig&0xf0) >> 4;
		}
		if (dig != searchstring[j]) { break; }
	    }
	    if (j == m) {
#ifdef LOOPCOUNT
		printf("eloop: %d  oloop: %d\n", eloop, oloop);
#endif
		return i;
	    }
	}
    }
    return i;
}

size_t search(unsigned char *buf, unsigned char *searchstring, size_t size, size_t start)
{
    int m;
    int res;
    int i;

    m = strlen(searchstring);

    for (i = 0; i < m; i++) {
	searchstring[i] -= '0';
    }
    
    if (m >= 5) {
	res = big_search(buf, searchstring, size, start, m);
    } else {
	res = sm_search(buf, searchstring, size, start, m);
    }
    
    for (i = 0; i < m; i++) {
	searchstring[i] += '0';
    }
    if (res == size*2) {
	return -1;
    }
    return res;
}
